@@ -34,6 +34,7 @@ import (
 	"github.com/rook/rook/pkg/operator/ceph/cluster/mon"
 	opconfig "github.com/rook/rook/pkg/operator/ceph/config"
 	opcontroller "github.com/rook/rook/pkg/operator/ceph/controller"
+	"github.com/rook/rook/pkg/operator/ceph/reporting"
 	"github.com/rook/rook/pkg/operator/k8sutil"
 	"github.com/rook/rook/pkg/util/exec"
 	appsv1 "k8s.io/api/apps/v1"
@@ -73,6 +74,9 @@ var controllerTypeMeta = metav1.TypeMeta{
 	APIVersion: fmt.Sprintf("%s/%s", cephv1.CustomResourceGroup, cephv1.Version),
 }
 
+// allow this to be overridden for unit tests
+var cephObjectStoreDependents = CephObjectStoreDependents
+
 // ReconcileCephObjectStore reconciles a cephObjectStore object
 type ReconcileCephObjectStore struct {
 	client              client.Client
@@ -82,6 +86,7 @@ type ReconcileCephObjectStore struct {
 	clusterSpec         *cephv1.ClusterSpec
 	clusterInfo         *cephclient.ClusterInfo
 	objectStoreChannels map[string]*objectStoreHealth
+	recorder            *k8sutil.EventReporter
 }
 
 type objectStoreHealth struct {
@@ -102,6 +107,7 @@ func newReconciler(mgr manager.Manager, context *clusterd.Context) reconcile.Rec
 	if err := cephv1.AddToScheme(mgr.GetScheme()); err != nil {
 		panic(err)
 	}
+
 	context.Client = mgr.GetClient()
 	return &ReconcileCephObjectStore{
 		client:              mgr.GetClient(),
@@ -109,6 +115,7 @@ func newReconciler(mgr manager.Manager, context *clusterd.Context) reconcile.Rec
 		context:             context,
 		bktclient:           bktclient.NewForConfigOrDie(context.KubeConfig),
 		objectStoreChannels: make(map[string]*objectStoreHealth),
+		recorder:            k8sutil.NewEventReporter(mgr.GetEventRecorderFor("rook-" + controllerName)),
 	}
 }
 
@@ -165,31 +172,28 @@ func add(mgr manager.Manager, r reconcile.Reconciler) error {
 // Result.Requeue is true, otherwise upon completion it will remove the work from the queue.
 func (r *ReconcileCephObjectStore) Reconcile(context context.Context, request reconcile.Request) (reconcile.Result, error) {
 	// workaround because the rook logging mechanism is not compatible with the controller-runtime logging interface
-	reconcileResponse, err := r.reconcile(request)
-	if err != nil {
-		logger.Errorf("failed to reconcile %v", err)
-	}
+	reconcileResponse, objectStore, err := r.reconcile(request)
 
-	return reconcileResponse, err
+	return reporting.ReportReconcileResult(logger, r.recorder, objectStore, reconcileResponse, err)
 }
 
-func (r *ReconcileCephObjectStore) reconcile(request reconcile.Request) (reconcile.Result, error) {
+func (r *ReconcileCephObjectStore) reconcile(request reconcile.Request) (reconcile.Result, *cephv1.CephObjectStore, error) {
 	// Fetch the cephObjectStore instance
 	cephObjectStore := &cephv1.CephObjectStore{}
 	err := r.client.Get(context.TODO(), request.NamespacedName, cephObjectStore)
 	if err != nil {
 		if kerrors.IsNotFound(err) {
 			logger.Debug("cephObjectStore resource not found. Ignoring since object must be deleted.")
-			return reconcile.Result{}, nil
+			return reconcile.Result{}, cephObjectStore, nil
 		}
 		// Error reading the object - requeue the request.
-		return reconcile.Result{}, errors.Wrap(err, "failed to get cephObjectStore")
+		return reconcile.Result{}, cephObjectStore, errors.Wrap(err, "failed to get cephObjectStore")
 	}
 
 	// Set a finalizer so we can do cleanup before the object goes away
 	err = opcontroller.AddFinalizerIfNotPresent(r.client, cephObjectStore)
 	if err != nil {
-		return reconcile.Result{}, errors.Wrap(err, "failed to add finalizer")
+		return reconcile.Result{}, cephObjectStore, errors.Wrap(err, "failed to add finalizer")
 	}
 
 	// The CR was just created, initializing status fields
@@ -199,7 +203,7 @@ func (r *ReconcileCephObjectStore) reconcile(request reconcile.Request) (reconci
 	}
 
 	// Make sure a CephCluster is present otherwise do nothing
-	cephCluster, isReadyToReconcile, cephClusterExists, reconcileResponse := opcontroller.IsReadyToReconcile(r.client, r.context, request.NamespacedName, controllerName)
+	cephCluster, isReadyToReconcile, cephClusterExists, reconcileResponse := opcontroller.IsReadyToReconcile(r.client, request.NamespacedName, controllerName)
 	if !isReadyToReconcile {
 		// This handles the case where the Ceph Cluster is gone and we want to delete that CR
 		// We skip the deleteStore() function since everything is gone already
@@ -211,14 +215,14 @@ func (r *ReconcileCephObjectStore) reconcile(request reconcile.Request) (reconci
 			// Remove finalizer
 			err := opcontroller.RemoveFinalizer(r.client, cephObjectStore)
 			if err != nil {
-				return reconcile.Result{}, errors.Wrap(err, "failed to remove finalizer")
+				return reconcile.Result{}, cephObjectStore, errors.Wrap(err, "failed to remove finalizer")
 			}
 
 			// Return and do not requeue. Successful deletion.
-			return reconcile.Result{}, nil
+			return reconcile.Result{}, cephObjectStore, nil
 		}
 
-		return reconcileResponse, nil
+		return reconcileResponse, cephObjectStore, nil
 	}
 	r.clusterSpec = &cephCluster.Spec
 
@@ -235,7 +239,7 @@ func (r *ReconcileCephObjectStore) reconcile(request reconcile.Request) (reconci
 	// Populate clusterInfo during each reconcile
 	r.clusterInfo, _, _, err = mon.LoadClusterInfo(r.context, request.NamespacedName.Namespace)
 	if err != nil {
-		return reconcile.Result{}, errors.Wrap(err, "failed to populate cluster info")
+		return reconcile.Result{}, cephObjectStore, errors.Wrap(err, "failed to populate cluster info")
 	}
 
 	// Populate CephVersion
@@ -243,34 +247,52 @@ func (r *ReconcileCephObjectStore) reconcile(request reconcile.Request) (reconci
 	if err != nil {
 		if strings.Contains(err.Error(), opcontroller.UninitializedCephConfigError) {
 			logger.Info(opcontroller.OperatorNotInitializedMessage)
-			return opcontroller.WaitForRequeueIfOperatorNotInitialized, nil
+			return opcontroller.WaitForRequeueIfOperatorNotInitialized, cephObjectStore, nil
 		}
-		return reconcile.Result{}, errors.Wrapf(err, "failed to retrieve current ceph %q version", opconfig.MonType)
+		return reconcile.Result{}, cephObjectStore, errors.Wrapf(err, "failed to retrieve current ceph %q version", opconfig.MonType)
 	}
 	r.clusterInfo.CephVersion = currentCephVersion
 
 	// DELETE: the CR was deleted
 	if !cephObjectStore.GetDeletionTimestamp().IsZero() {
-		logger.Debugf("deleting store %q", cephObjectStore.Name)
+		updateStatus(r.client, request.NamespacedName, cephv1.ConditionDeleting, buildStatusInfo(cephObjectStore))
 
 		if ok {
-			// Close the channel to stop the healthcheck of the endpoint
-			close(r.objectStoreChannels[cephObjectStore.Name].stopChan)
-
-			response, okToDelete := r.verifyObjectBucketCleanup(cephObjectStore)
-			if !okToDelete {
-				// If the object store cannot be deleted, requeue the request for deletion to see if the conditions
-				// will eventually be satisfied such as the object buckets being removed
-				return response, nil
+			select {
+			case <-r.objectStoreChannels[cephObjectStore.Name].stopChan:
+				// channel was closed
+				break
+			default:
+				// Close the channel to stop the healthcheck of the endpoint
+				close(r.objectStoreChannels[cephObjectStore.Name].stopChan)
 			}
 
-			response, okToDelete = r.verifyObjectUserCleanup(cephObjectStore)
-			if !okToDelete {
-				// If the object store cannot be deleted, requeue the request for deletion to see if the conditions
-				// will eventually be satisfied such as the object users being removed
-				return response, nil
+			// get the latest version of the object now that the health checker is stopped
+			err := r.client.Get(context.TODO(), request.NamespacedName, cephObjectStore)
+			if err != nil {
+				return reconcile.Result{}, cephObjectStore, errors.Wrapf(err, "failed to get latest CephObjectStore %q", request.NamespacedName.String())
 			}
 
+			objCtx, err := NewMultisiteContext(r.context, r.clusterInfo, cephObjectStore)
+			if err != nil {
+				return reconcile.Result{}, cephObjectStore, errors.Wrapf(err, "failed to check for object buckets. failed to get object context")
+			}
+
+			opsCtx, err := NewMultisiteAdminOpsContext(objCtx, &cephObjectStore.Spec)
+			if err != nil {
+				return reconcile.Result{}, cephObjectStore, errors.Wrapf(err, "failed to check for object buckets. failed to get admin ops API context")
+			}
+
+			deps, err := cephObjectStoreDependents(r.context, r.clusterInfo, cephObjectStore, objCtx, opsCtx)
+			if err != nil {
+				return reconcile.Result{}, cephObjectStore, err
+			}
+			if !deps.Empty() {
+				err := reporting.ReportDeletionBlockedDueToDependents(logger, r.client, cephObjectStore, deps)
+				return opcontroller.WaitForRequeueIfFinalizerBlocked, cephObjectStore, err
+			}
+			reporting.ReportDeletionNotBlockedDueToDependents(logger, r.client, r.recorder, cephObjectStore)
+
 			cfg := clusterConfig{
 				context:     r.context,
 				store:       cephObjectStore,
@@ -286,16 +308,16 @@ func (r *ReconcileCephObjectStore) reconcile(request reconcile.Request) (reconci
 		// Remove finalizer
 		err = opcontroller.RemoveFinalizer(r.client, cephObjectStore)
 		if err != nil {
-			return reconcile.Result{}, errors.Wrap(err, "failed to remove finalizer")
+			return reconcile.Result{}, cephObjectStore, errors.Wrap(err, "failed to remove finalizer")
 		}
 
 		// Return and do not requeue. Successful deletion.
-		return reconcile.Result{}, nil
+		return reconcile.Result{}, cephObjectStore, nil
 	}
 
 	// validate the store settings
 	if err := r.validateStore(cephObjectStore); err != nil {
-		return reconcile.Result{}, errors.Wrapf(err, "invalid object store %q arguments", cephObjectStore.Name)
+		return reconcile.Result{}, cephObjectStore, errors.Wrapf(err, "invalid object store %q arguments", cephObjectStore.Name)
 	}
 
 	// If the CephCluster has enabled the "pg_autoscaler" module and is running Nautilus
@@ -314,9 +336,10 @@ func (r *ReconcileCephObjectStore) reconcile(request reconcile.Request) (reconci
 	_, err = r.reconcileCreateObjectStore(cephObjectStore, request.NamespacedName, cephCluster.Spec)
 	if err != nil && kerrors.IsNotFound(err) {
 		logger.Info(opcontroller.OperatorNotInitializedMessage)
-		return opcontroller.WaitForRequeueIfOperatorNotInitialized, nil
+		return opcontroller.WaitForRequeueIfOperatorNotInitialized, cephObjectStore, nil
 	} else if err != nil {
-		return r.setFailedStatus(request.NamespacedName, "failed to create object store deployments", err)
+		result, err := r.setFailedStatus(request.NamespacedName, "failed to create object store deployments", err)
+		return result, cephObjectStore, err
 	}
 
 	// Set Progressing status, we are done reconciling, the health check go routine will update the status
@@ -324,7 +347,7 @@ func (r *ReconcileCephObjectStore) reconcile(request reconcile.Request) (reconci
 
 	// Return and do not requeue
 	logger.Debug("done reconciling")
-	return reconcile.Result{}, nil
+	return reconcile.Result{}, cephObjectStore, nil
 }
 
 func (r *ReconcileCephObjectStore) reconcileCreateObjectStore(cephObjectStore *cephv1.CephObjectStore, namespacedName types.NamespacedName, cluster cephv1.ClusterSpec) (reconcile.Result, error) {
@@ -362,6 +385,10 @@ func (r *ReconcileCephObjectStore) reconcileCreateObjectStore(cephObjectStore *c
 		if err != nil {
 			return r.setFailedStatus(namespacedName, "failed to reconcile external endpoint", err)
 		}
+
+		if err := UpdateEndpoint(objContext, &cephObjectStore.Spec); err != nil {
+			return r.setFailedStatus(namespacedName, "failed to set endpoint", err)
+		}
 	} else {
 		logger.Info("reconciling object store deployments")
 
@@ -391,6 +418,10 @@ func (r *ReconcileCephObjectStore) reconcileCreateObjectStore(cephObjectStore *c
 			return r.setFailedStatus(namespacedName, "failed to reconcile service", err)
 		}
 
+		if err := UpdateEndpoint(objContext, &cephObjectStore.Spec); err != nil {
+			return r.setFailedStatus(namespacedName, "failed to set endpoint", err)
+		}
+
 		// Reconcile Pool Creation
 		if !cephObjectStore.Spec.IsMultisite() {
 			logger.Info("reconciling object store pools")
@@ -418,7 +449,10 @@ func (r *ReconcileCephObjectStore) reconcileCreateObjectStore(cephObjectStore *c
 
 	// Start monitoring
 	if !cephObjectStore.Spec.HealthCheck.Bucket.Disabled {
-		r.startMonitoring(cephObjectStore, objContext, namespacedName)
+		err = r.startMonitoring(cephObjectStore, objContext, namespacedName)
+		if err != nil {
+			return reconcile.Result{}, err
+		}
 	}
 
 	return reconcile.Result{}, nil
@@ -483,84 +517,23 @@ func (r *ReconcileCephObjectStore) reconcileMultisiteCRs(cephObjectStore *cephv1
 	return cephObjectStore.Name, cephObjectStore.Name, cephObjectStore.Name, reconcile.Result{}, nil
 }
 
-func (r *ReconcileCephObjectStore) verifyObjectBucketCleanup(objectstore *cephv1.CephObjectStore) (reconcile.Result, bool) {
-	bktProvisioner := GetObjectBucketProvisioner(r.context, objectstore.Namespace)
-	bktProvisioner = strings.Replace(bktProvisioner, "/", "-", -1)
-	selector := fmt.Sprintf("bucket-provisioner=%s", bktProvisioner)
-	objectBuckets, err := r.bktclient.ObjectbucketV1alpha1().ObjectBuckets().List(context.TODO(), metav1.ListOptions{LabelSelector: selector})
-	if err != nil {
-		logger.Errorf("failed to delete object store. failed to list buckets for objectstore %q in namespace %q", objectstore.Name, objectstore.Namespace)
-		return opcontroller.WaitForRequeueIfFinalizerBlocked, false
-	}
-
-	if len(objectBuckets.Items) == 0 {
-		logger.Infof("no buckets found for objectstore %q in namespace %q", objectstore.Name, objectstore.Namespace)
-		return reconcile.Result{}, true
-	}
-
-	bucketNames := make([]string, 0)
-	for _, bucket := range objectBuckets.Items {
-		bucketNames = append(bucketNames, bucket.Name)
-	}
-
-	logger.Errorf("failed to delete object store. buckets for objectstore %q in namespace %q are not cleaned up. remaining buckets: %+v", objectstore.Name, objectstore.Namespace, bucketNames)
-	return opcontroller.WaitForRequeueIfFinalizerBlocked, false
-}
-
-func (r *ReconcileCephObjectStore) startMonitoring(objectstore *cephv1.CephObjectStore, objContext *Context, namespacedName types.NamespacedName) {
+func (r *ReconcileCephObjectStore) startMonitoring(objectstore *cephv1.CephObjectStore, objContext *Context, namespacedName types.NamespacedName) error {
 	// Start monitoring object store
 	if r.objectStoreChannels[objectstore.Name].monitoringRunning {
-		logger.Debug("external rgw endpoint monitoring go routine already running!")
-		return
+		logger.Info("external rgw endpoint monitoring go routine already running!")
+		return nil
 	}
 
+	rgwChecker, err := newBucketChecker(r.context, objContext, r.client, namespacedName, &objectstore.Spec)
+	if err != nil {
+		return errors.Wrapf(err, "failed to start rgw health checker for CephObjectStore %q, will re-reconcile", namespacedName.String())
+	}
+
+	logger.Infof("starting rgw health checker for CephObjectStore %q", namespacedName.String())
+	go rgwChecker.checkObjectStore(r.objectStoreChannels[objectstore.Name].stopChan)
+
 	// Set the monitoring flag so we don't start more than one go routine
 	r.objectStoreChannels[objectstore.Name].monitoringRunning = true
 
-	var port int32
-
-	if objectstore.Spec.IsTLSEnabled() {
-		port = objectstore.Spec.Gateway.SecurePort
-	} else if objectstore.Spec.Gateway.Port != 0 {
-		port = objectstore.Spec.Gateway.Port
-	} else {
-		logger.Error("At least one of Port or SecurePort should be non-zero")
-		return
-	}
-
-	rgwChecker := newBucketChecker(r.context, objContext, port, r.client, namespacedName, &objectstore.Spec)
-
-	// Fetch the admin ops user
-	accessKey, secretKey, err := GetAdminOPSUserCredentials(r.context, r.clusterInfo, objContext, objectstore)
-	if err != nil {
-		logger.Errorf("failed to create or retrieve rgw admin ops user. %v", err)
-		return
-	}
-	rgwChecker.objContext.adminOpsUserAccessKey = accessKey
-	rgwChecker.objContext.adminOpsUserSecretKey = secretKey
-
-	logger.Info("starting rgw healthcheck")
-	go rgwChecker.checkObjectStore(r.objectStoreChannels[objectstore.Name].stopChan)
-}
-
-func (r *ReconcileCephObjectStore) verifyObjectUserCleanup(objectstore *cephv1.CephObjectStore) (reconcile.Result, bool) {
-	ctx := context.TODO()
-	cephObjectUsers, err := r.context.RookClientset.CephV1().CephObjectStoreUsers(objectstore.Namespace).List(ctx, metav1.ListOptions{})
-	if err != nil {
-		logger.Errorf("failed to delete object store. failed to list user for objectstore %q in namespace %q", objectstore.Name, objectstore.Namespace)
-		return opcontroller.WaitForRequeueIfFinalizerBlocked, false
-	}
-
-	if len(cephObjectUsers.Items) == 0 {
-		logger.Infof("no users found for objectstore %q in namespace %q", objectstore.Name, objectstore.Namespace)
-		return reconcile.Result{}, true
-	}
-
-	userNames := make([]string, 0)
-	for _, user := range cephObjectUsers.Items {
-		userNames = append(userNames, user.Name)
-	}
-
-	logger.Errorf("failed to delete object store. users for objectstore %q in namespace %q are not cleaned up. remaining users: %+v", objectstore.Name, objectstore.Namespace, userNames)
-	return opcontroller.WaitForRequeueIfFinalizerBlocked, false
+	return nil
 }
